# world.py - represent the game world

import Player
import Route


Clear = '.'
Base  = '@'
Wall  = '#'
Water = '~'

North =  1
South = -1
East  =  1
West  = -1

_dir_map = { 'N' : (0, 1), 'S' : (0, -1), 'E' : (1, 0), 'W' : (-1, 0) }
_world = None

def create_world(worldmap, playerdata):
    global _world
    if not _world:
        _world = World(worldmap)
        _world.set_us(Player.Player(playerdata[0], playerdata[1], playerdata[2]))
    return _world

def get_world():
    return _world


class Package:
    def __init__(self, id):
        self.id = id
        self._weight = 0
        self._x, self._y = None, None
        self._xdest, self._ydest = None, None
        self._dropped = 0

    def drop(self):
        self._dropped = 1

    def pick(self):
        self._dropped = 0

    def set_destination(self, x, y):
        self._xdest = x
        self._ydest = y

    def set_position(self, x, y):
        self._x = x
        self._y = y

    def set_weight(self, weight):
        self._weight = weight

    def destination(self):
        return (self._xdest, self._ydest)

    def position(self):
        return (self._x, self._y)

    def weight(self):
        return self._weight

    def possibly_delivered(self):
        return self._dropped


class World:
    def __init__(self, worldmap):
        self.robot = None
        # Don't ask.
        self._map = map(list, apply(map, (None,) + tuple(map(list, worldmap))))
        self._width = len(self._map)
        self._height = len(self._map[0])
        self._bases = []   #SJK
        self._find_bases() #SJK
        self._packages = {}
        self._robots = {}

    def _find_bases(self):
        for y in range(self._height):
            for x in range(self._width):
                if self._map[x][y] == Base:
                    self._bases.append((x+1,y+1))

        
    def set_us(self, bot):
        self.robot = bot
        self._robots[bot.id] = bot

    def get_bot(self, bot_id, create=0):
        bot = self._robots.get(bot_id)
        if not bot and create:
            bot = Player.Robot(bot_id)
            self._robots[bot_id] = bot
        return bot

    # Kill?
    def get_map(self):
        return self._map


    def _get_route(self, x, y, dx, dy):
        route = Route.Route(dx, dy)
        return route.find_path(x, y)

    def _get_simple_distance(self, x, y, dx, dy):
        return (abs(x-dx)+abs(y-dy), dx,dy) 

    def _get_optimal_distance(self,x,y,dx,dy):
        route = self._get_route(x,y,dx,dy)
        return (len(route), route) 
      
    def get_next_package(self,x,y):
        dists = []
        for dx, dy in self._packages.values():
            dists.append(self._get_optimal_distance(x,y,dx,dy))
        dists.sort()
        if dists:
            if isinstance(dists[0][1], list):   
                return dists[0][1]
            else: 
                dx, dx = dists[0][1]
                return self._get_route(x,y,dx,dy)
        return []

    def get_next_base(self, x, y):
        dists = []
        for dx, dy in self._bases:
            if (x, y) != (dx, dy):
                dists.append(self._get_optimal_distance(x,y,dx,dy))
        dists.sort()
        if dists:
            if isinstance(dists[0][1], list):   
                return dists[0][1]
            else: 
                dx, dx = dists[0][1]
                return self._get_route(x,y,dx,dy)
        return []


##    def get_pkgs(self):
##        pkgs = []
##        for pkg in self._packages:
##            pkgs.append( (pkg.weight(), pkg.position()) )
##        for base in self._bases:
##            pkgs.append( (-1, base) )
##        return pkgs 

    def get_pkg(self, pkg_id, create=0):
        pkg = self._packages.get(pkg_id)
        if not pkg and create:
            pkg = Package(pkg_id)
            self._packages[pkg_id] = pkg
        return pkg

    def packages_at(self, x, y):
        return [ p for p in self._packages.values() if p.position() == (x, y) ]

    def remove_package(self, pkg):
        if self._packages.has_key(pkg.id):
            del self._packages[pkg.id]
            print "Deleted Pkg:", pkg.id

    def whats_at(self, x, y):
        if x <= 0 or x > self._width:
            return Wall
        if y <= 0 or y > self._height:
            return Wall
        # Adjust for map origin of (1,1).
        return self._map[x-1][y-1]

    def get_adjacent(self,x,y):
        adjs = []
        for dir in ('N', 'S', 'E', 'W'):
            px, py = _dir_map[dir]
            px += x
            py += y
            obj = self.whats_at(px, py)
            bots = [ r for r in self._robots.values() if r.position() == (px, py) ]
            if (obj == Clear or obj == Base) and bots:
                obj = None         
            adjs.append(obj)
        return adjs

    def update_packages(self, packages):
        x, y = self.robot.position()
        if packages:
            for id, dx, dy, weight in packages:
                pkg = self.get_pkg(id, 1)
                pkg.set_position(x, y)
                pkg.set_destination(dx, dy)
                pkg.set_weight(weight)
        else:
            try: self._bases.remove((x, y))
            except ValueError: pass
            for package in self._packages.values():
                if (x, y) == package.position() and package.possibly_delivered():
                    self.remove_package(package)

    def update(self, srv_updates):
        bots_seen = []
        for bot_actions in filter(None, srv_updates.split('#')):
            action_list = bot_actions.split()
            try:
                bot_id = int(action_list.pop(0))
                bots_seen.append(bot_id)
                bot = self.get_bot(bot_id, 1)
                print "Bot:", bot_id,

                while action_list:
                    action = action_list.pop(0).upper()
                    print "action", action,

                    if action in ('N', 'S', 'E', 'W'):
                        dx, dy = _dir_map[action]
                        bot.move_by(dx, dy)
                        print "moved bot by", (dx, dy),
                    elif action == 'P':
                        bot.pick_pkg(self.get_pkg(int(action_list.pop(0)), 1))
                    elif action == 'D':
                        bot.drop_pkg(self.get_pkg(int(action_list.pop(0))))
                    elif action == 'X':
                        abs_x = int(action_list.pop(0))
                        action_list.pop(0)  # Remove the 'Y'
                        abs_y = int(action_list.pop(0))
                        bot.set_position(abs_x, abs_y)
                        "moved bot to", (abs_x, abs_y),
                print
                
            except (ValueError, KeyError, IndexError):
                pass
        all_bots = self._robots.keys()
        try:
            for bot_id in bots_seen:
                all_bots.remove(bot_id)
            for bot_id in all_bots:
                bot = self.get_bot(bot_id)
                if bot:
                    bot.die()
                    del self._robots[bot_id]
        except ValueError:
            print "Saw bot #%d but not in _robots" % bot_id

    def print_map(self, space=' ', path=None):
        print
        for y in range(self._height):
            line = ''
            for x in range(self._width):
                out = self._map[x][y]
                if path and (x+1, y+1) in path:
                    out = 'o'
                line += out + space
            print line

